/* -*- linux-c -*- */
/*
 * plx9080.c
 *
 * Specific definitions and functions for the PCI9080 used by plxpci
 * Written by: niko.neufeld@cern.ch CERN/LHCb 2002
 */

/* $Id: plx9080.c,v 1.1 2002/10/01 12:22:35 nikolaus Exp $ */

#include <linux/types.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <asm/system.h>
#include <asm/hardirq.h>
#include <asm/io.h>

#include "plxtypes.h"
#include "plx9080.h"
#include "plx9080reg.h"

static struct plx_ops plx9080_ops __attribute__ ((unused)) = {
	&plx9080_init,
	&plx9080_init_local_bus,
	&plx_irqh,
	&plx_read_local_bus,
	&plx_write_local_bus,
	&plx9080_cleanup
};

int plx9080_init(struct plx_dev *dev)
{
	u32 w;

	printk("%s: initialising\n", dev->name);
	plx_read_config_dword(dev, PCI9080_EEPROM_CTRL_STAT, &w);
	w &= ~(10000); /* set user output to 0 */
	plx_write_config_dword(dev, PCI9080_EEPROM_CTRL_STAT, w);	
	return 0;
}

/*
 * activate and map local bus space 0 into the host memory
 * autoprobing of suitable memory region works only in 2.4.x :-(
 */
int plx9080_init_local_bus(struct plx_dev *dev, plx_loc_mem_t *memdef)
{
	struct pci_dev *pdev = dev->pci_dev;
	u32 mask, base_addr;
	unsigned long size;
	u32 cur, w;
	int status = 0;

	printk("%s: initialising\n", dev->name);
	plx_read_config_dword(dev, PCI9080_EEPROM_CTRL_STAT, &w);
	w &= ~(10000); /* set user output to 0 */
	plx_write_config_dword(dev, PCI9080_EEPROM_CTRL_STAT, w);
	
	mask = memdef->address_mask;
	size = (~memdef->address_mask + 1);
	mask &= ~0xF;     /* non prefetchable memory anywhere in 32bit space */
	printk(KERN_INFO "%s: try to bus 0 decode %08x (range of %lx bytes)\n",
	       dev->name, mask, size);
	mb();
	cli();
	plx_write_config_dword(dev, PCI9080_SPACE0_RANGE, mask);
	/* enable space 0 at offset 0 */
	wmb();
	plx_write_config_dword(dev, PCI9080_SPACE0_REMAP, 0x1);
	/* probe the size we have to map */
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_2, &cur);
	pci_write_config_dword(pdev, PCI_BASE_ADDRESS_2, ~0);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_2, &mask);
	pci_write_config_dword(pdev, PCI_BASE_ADDRESS_2, cur);
	sti();
	if (mask & PCI_BASE_ADDRESS_SPACE_IO) {
		printk("%s: local bus maps into I/O space - no good!\n",
		       dev->name);
	        status = -EBUSY;
		goto error;
	} else {
		mask &= PCI_BASE_ADDRESS_MEM_MASK;
		size = (~mask + 1);
	}
	printk("%s: attempting to reserve region of 0x%08lx bytes; mask 0x%08x\n",
	       dev->name, size, mask);
#if (LINUX_VERSION_CODE >= 0x020400)
	if ((memdef->base_address = find_mem_region(&size)) == NULL) {
		sti();
		printk("%s: found no memory region of 0x%08x bytes\n",
		       dev->name, size);
		status = -EBUSY; 
		goto error;
	}
	request_mem_region(memdef->base_address, size, dev->name); /* can't fail */
#endif
	if (memdef->base_address == 0) {
	    printk("%s: invalid memory region for remap 0x%08x\n",
		   dev->name, memdef->base_address);
	    goto error;
	}
	printk("%s: Trying to remap local space 0 to 0x%08x\n", 
	       dev->name, memdef->base_address);
	cli();
	mb();
	pci_write_config_dword(pdev, PCI_BASE_ADDRESS_2, memdef->base_address);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_2, &base_addr);
	sti();
	printk("%s: mapped local bus 0 to base 0x%08x\n", dev->name, 
	       base_addr);
	if (!(dev->baseaddr[0] = (u32) ioremap(base_addr, size))) {
		printk("%s: could not remap local bus 0 area\n", dev->name);
		status = -ENOMEM;
		goto error;
	}
	printk("%s: remapped local bus 0 to virtual address 0x%08x\n",
	       dev->name, dev->baseaddr[0]);
	dev->winsize[0] = size;
	return 0;

error:	
	if (dev->baseaddr[0]) iounmap((unsigned long *)dev->baseaddr[0]);
	dev->baseaddr[0] = dev->winsize[0] = 0;
	wmb();
	plx_write_config_dword(dev, PCI9080_SPACE0_REMAP, 0x0);
	plx_write_config_dword(dev, PCI9080_SPACE0_RANGE, 0x0);
	return status;

}
	
void plx9080_cleanup(struct plx_dev *dev) 
{
	synchronize_irq();
	if (dev->baseaddr[0]) iounmap((unsigned long *) dev->baseaddr[0]);
#if (LINUX_VERSION_CODE >= 0x020400)
	if (dev->baseaddr[0]) release_mem_region(dev->baseaddr[0], 
						 dev->winsize[0]);
#endif
	printk("%s unmapping bus 0\n", dev->name);
	plx_write_config_dword(dev, PCI9080_SPACE0_REMAP, 0x0);
}














