/* -*- linux-c -*- */
/* 
 * plxpci.c
 *
 * Simple-minded driver for access to basic plx pci90xx devices functionality
 * It draws on the plx-sdk-0.2.0 for linux 
 * Written by: niko.neufeld@cern.ch CERN/LHCb 2002 
 */

/* $Id: plxpci.c,v 1.2 2002/10/01 16:16:46 nikolaus Exp $ */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>       /* printk() */
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/ioport.h>       /* resource */
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/sched.h>
#if (LINUX_VERSION_CODE < 0x020400)
#include <asm/spinlock.h>  
#include <linux/malloc.h>       /* kmalloc() */
#else
#include <linux/spinlock.h>
#include <linux/slab.h>
#endif
#include <asm/io.h>        /*  ioremap */
#include <asm/system.h>    /*  mb() */
#include <asm/uaccess.h>   /*  copy_to_user */
#include <asm/byteorder.h> /*  efficient byte swapping */
#include <asm/hardirq.h>
#include <linux/ioctl.h> 

#include "Plx.h"
#include "plxpci.h"
#include "plxioctl.h"
#include "plx9030.h"
#include "plx9080.h"

static char version[] __initdata = 
  "plx.c: $Id: plxpci.c,v 1.2 2002/10/01 16:16:46 nikolaus Exp $\n"
  "       Niko Neufeld, CERN/LHCb niko.neufeld@cern.ch\n";

static int probed __initdata = 0; 
static int num_plx = 0;
static struct plx_dev *plx_device[PLX_MAX_DEVICE] = {0,};
static int plx_major = PLX_MAJOR;

#if (LINUX_VERSION_CODE < 0x020400)
static struct file_operations plx_fops =
{
        NULL,           /* lseek */
        NULL,           /* read */
        NULL,           /* write */
        NULL,           /* readdir */
        NULL,           /* poll */
        &plx_ioctl,     /* ioctl */
        NULL,           /* mmap */
        &plx_open,      /* open */
        NULL,           /* flush */
        &plx_release,   /* release */
        NULL,           /* fsync */
        NULL,           /* fasync */
        NULL,           /* check_media_change */
        NULL,           /* revalidate */
        NULL            /* lock */
};
#else /* kernen 2.4.x and higher */
static struct file_operations plx_fops =
{
        NULL,           /* owner */
        NULL,           /* lseek */ 
        NULL,           /* read */
        NULL,           /* write */
        NULL,           /* readdir */
        NULL,           /* poll */
        &plx_ioctl,
        NULL,           /* mmap */
        &plx_open,
        NULL,           /* flush */
        &plx_release,
        NULL,           /* fsync */
        NULL,           /* fasync */
        NULL,           /* lock */
        NULL,           /* readv */
        NULL,           /* writev */
        NULL            /* writepage */    
};

#endif /* if LINUX_VERSION_CODE < 0x020400 */

static struct plx_ops plx_default_ops = {
	&plx_init,
	&plx_init_local_bus,
	&plx_irqh,
	&plx_read_local_bus,
	&plx_write_local_bus,
	&plx_cleanup
};

int plx_open(struct inode *inode, struct file *filp) 
{  
        struct plx_dev *dev;
        int num;
        
        if ((num = MINOR(inode->i_rdev)) >= num_plx) 
                return -ENODEV;
	if (!plx_device[num]) {
		printk(KERN_INFO "plx: minor %d not initialised\n", num);
		return -EBUSY;
	}
        dev = plx_device[num];
        filp->private_data = (void *) dev;
        printk(KERN_INFO "%s opened\n", dev->name);
        MOD_INC_USE_COUNT;  
        return 0; 
}

int plx_release(struct inode *inode, struct file * filp) 
{
        int num;
        struct plx_dev *dev;
        
        if ( (num = MINOR(inode->i_rdev)) >= num_plx) 
                return -ENODEV;
        dev = (struct plx_dev *) filp->private_data;
	printk(KERN_INFO "%s closed\n", dev->name);
        MOD_DEC_USE_COUNT;
        return 0;
}
  
int plx_ioctl(struct inode *ip, struct file *filp, unsigned int cmd,
	      unsigned long arg) 
{
        struct plx_dev *dev = (struct plx_dev *) filp->private_data;
	int size = _IOC_SIZE(cmd), err = 0; 
	struct plx_reg *regp;
	struct plx_loc_mem *locmemp;
	struct plx_loc_data *locdatp;
        
        if ((_IOC_TYPE(cmd)) != PLX_IOC_MAGIC) return -EINVAL;
        if (_IOC_NR(cmd) > PLX_IOC_MAXNR) return -EINVAL;
        
        if (_IOC_DIR(cmd) & _IOC_READ)
                err = verify_area(VERIFY_WRITE, (void *) arg, size);
        else if (_IOC_DIR(cmd) & _IOC_WRITE)
                err = verify_area(VERIFY_READ, (void *) arg, size);
        if (err) return err;

	switch (cmd) {
	case PLX_INIT_LOC_BUS:
		locmemp = (struct plx_loc_mem *) arg;
		/* temporary hack to avoid double initialisation */
		/* probably should return an error here */
		if ((dev->baseaddr[0])) return 0;
		if ((err = (*(dev->ops)->init_local_bus)(dev, locmemp)) < 0) {
			printk(KERN_WARNING "plx: %s failed to initialise"\
			       " local bus\n", dev->name);
			return err;
		}
		break;
	case PLX_READ_LOC_BUS:
		locdatp = (struct plx_loc_data *) arg;
		if (!(dev->ops)->read_local_bus) {
		     printk("plx: function pointer corruption!\n");
		     return -EINVAL;
		}
		if ((*(dev->ops)->read_local_bus)(dev, locdatp)) {
			return -ENOMEM;
		}
		break;
	case PLX_WRITE_LOC_BUS:
		locdatp = (struct plx_loc_data *) arg;
		if (!(dev->ops)->write_local_bus) {
		     printk("plx: function pointer corruption!\n");
		     return -EINVAL;
		}
		if ((*(dev->ops)->write_local_bus)(dev, locdatp)) {
			return -ENOMEM;
		}
		break;
	case PLX_READ_REG:
		regp = (struct plx_reg *) arg;
		if (regp->offset & 3) return -EINVAL;
		plx_read_config_dword(dev, regp->offset, &regp->value);
		break;
	case PLX_WRITE_REG:
		regp = (struct plx_reg *) arg;
		if (regp->offset & 3) return -EINVAL;
		plx_write_config_dword(dev, regp->offset, regp->value);
		break;
	default:
		printk("plx: %s unknown command %d\n", dev->name, cmd);
		return -EINVAL;
	}
	return 0;
}

struct plx_dev *plx_dev_init(struct pci_dev *pdev, struct plx_ops *pops) 
{
	struct plx_dev *plxp;
	unsigned int baseaddr = 0;

	if (!(plxp = (struct plx_dev *) kmalloc(sizeof(struct plx_dev), 
						GFP_KERNEL))) {
		goto error;
	}
	memset(plxp, 0, sizeof(struct plx_dev));
	plxp->pci_dev = pdev;
	plxp->irq = pdev->irq;
	plxp->ops = pops;

        PLX_DEVICE_SET_NAME(plxp->name, num_plx);
	
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0, &baseaddr);
	plxp->irq = pdev->irq;
	if (baseaddr == 0) {
		printk("plx: warning baseaddr 0 disabled\n");
		goto error;
	}
	if (!(plxp->regs = (unsigned long) 
	      ioremap(baseaddr, 0x100))) {
		goto error;
	}
	if (plxp->irq >= 0) {
		if (request_irq(plxp->irq, (plxp->ops)->irqh, SA_SHIRQ, 
				plxp->name, plxp)) {
			printk("plx: %s could not assign irq %d\n", 
			       plxp->name, plxp->irq);
			plxp->irq = -1;
		}
	}

	return (plxp);

error:
	if (plxp) kfree(plxp);
	if (plxp->regs) iounmap((unsigned long *) plxp->regs);
	return NULL;

}

int plx_probe(void)
{
	struct pci_dev *pdev;
	int status = 0;
	struct plx_ops *pops;

	if (probed)
		return -ENODEV;
	probed++;
	
	if (!pci_present())
		return -ENODEV;
#if (LINUX_VERSION_CODE < 0x020400)
	pdev = pci_devices; /* need to check all PCI devices */
	while (pdev) {
#else
	pci_for_each_dev(pdev) {
#endif
		if (pdev->vendor == PCI_VENDOR_ID_PLX ||
		    pdev->vendor == PCI_VENDOR_ID_ALTERA) {
			switch (pdev->device) {
			case PCI_DEVICE_ID_PLX_9030: 
			case PCI_DEVICE_ID_ALTERA_ZBIG:
				printk(KERN_INFO "plx: found 9030 device\n");
				pops = &plx9030_ops;
				break;
			case PCI_DEVICE_ID_PLX_9080:
				printk(KERN_INFO "plx: found 9080 device\n");
				pops = &plx9080_ops;
				break;
			default:
				printk(KERN_INFO"plx: found generic device"
				       " id %0x4x\n", pdev->device);
				pops = &plx_default_ops;
				break;
			}
			plx_device[num_plx] = plx_dev_init(pdev, pops);
			num_plx++;
		}
#if (LINUX_VERSION_CODE < 0x020400)
		pdev = pdev->next;
#endif 
	} /* while (pdev); */
	if (num_plx == 1) 
		printk(KERN_INFO "plx: found %d plx device\n", num_plx);
	else
		printk(KERN_INFO "plx: found %d plx devices\n", num_plx);
	
	return status;
}
				       

#ifdef MODULE
MODULE_AUTHOR("Niko Neufeld <niko.neufeld@cern.ch>");
/*MODULE_LICENSE("GPL");*/
MODULE_DESCRIPTION("PLX PCI-90xx driver");
#endif

void plx_module_cleanup(void)
{
	int i;
	struct plx_dev *dev;

	for (i = 0; i < PLX_MAX_DEVICE; ++i) {
		if (!plx_device[i]) continue;
		dev = plx_device[i];
		((dev->ops)->cleanup)(dev);
		synchronize_irq();
		if (dev->irq) free_irq(dev->irq, dev);
		if (dev->regs) iounmap((unsigned long *) dev->regs);
		kfree(dev);
	}
	unregister_chrdev(plx_major, "plx");
	printk(KERN_INFO "plxpci: driver unloaded\n");
}


int __init plx_module_init(void)
{
	int status;

	printk(KERN_INFO "\n%s", version);
	status = register_chrdev(plx_major, "plx", &plx_fops);
        if (status < 0) {
                printk(KERN_WARNING "plx: can't get major %d\n", 
                       status);
                return status;
        }
        if (plx_major == 0) plx_major = status;
	
	return plx_probe();
}


#if (LINUX_VERSION_CODE < 0x02032a)
#ifdef MODULE
int __init init_module(void)
{
        return plx_module_init();
}


void cleanup_module(void)
{
        plx_module_cleanup();
}
#endif /* ifdef MODULE */
#else
module_init(plx_module_init);
module_exit(plx_module_cleanup);
#endif /* (LINUX_VERSION_CODE < 0x02032a) */

int plx_write_config_dword(struct plx_dev *dev, u8 where, u32 val) 
{
	writel(val, dev->regs + where);
	return 0;
}

int plx_read_config_dword(struct plx_dev *dev, u8 where, u32 *val)
{
	*val = readl(dev->regs + where);
	return 0;
}

int plx_init(struct plx_dev *plxp) {
	return 0;
}

int plx_write_local_bus(struct plx_dev *dev, struct plx_loc_data *dat)
{
	u32 width = dat->flag & PLX_DATA_WIDTH_MASK;
	
/*	printk("%s: attempt to write %08x to %08x\n",
	       dev->name, dat->data, dev->baseaddr[0] + dat->address);
*/
	if (dev->baseaddr[0]) {
		switch (width) {
			u8 byte; u16 word;
		case PLX_DATA_WIDTH_LWORD:
			break; /* not implemented on 32bit architecture */
		case PLX_DATA_WIDTH_DWORD:
			writel(dat->data, dev->baseaddr[0] + dat->address);
			break;
		case PLX_DATA_WIDTH_WORD:
			word = dat->data;
			writew(word, dev->baseaddr[0] + dat->address);
			break;
		case PLX_DATA_WIDTH_BYTE:
			byte = dat->data;
			writeb(byte, dev->baseaddr[0] + dat->address);
			break;
		}
	} else {
		printk("%s: baseaddress invalid\n", dev->name);
		return -1;
	}
	return 0;
}

int plx_read_local_bus(struct plx_dev *dev, struct plx_loc_data *dat)
{
	u32 width = dat->flag & PLX_DATA_WIDTH_MASK;
	
//	printk(KERN_INFO "%s: read dword from address 0x%08lx\n", dev->name, 
//	       dev->baseaddr[0] + dat->address);

	if (dev->baseaddr[0]) {
		switch (width) {
		case PLX_DATA_WIDTH_LWORD:
			break; /* not implemented on 32bit architecture */
		case PLX_DATA_WIDTH_DWORD:
			dat->data = readl(dev->baseaddr[0] + dat->address);
			break;
		case PLX_DATA_WIDTH_WORD:
			dat->data = readw(dev->baseaddr[0] + dat->address);
			break;
		case PLX_DATA_WIDTH_BYTE:
			dat->data = readb(dev->baseaddr[0] + dat->address);
			break;
		}
	} else {
		printk("%s: baseaddress invalid\n", dev->name);
		return -1;
	}
	return 0;
}

void plx_cleanup(struct plx_dev *dev) {
	printk("%s: generic cleanup doing nothing\n", dev->name);
	return;
}

int plx_init_local_bus(struct plx_dev *plxp, plx_loc_mem_t *memt)
{
	printk("%s need specific plx to map local bus\n", plxp->name);
	return -1;
}

void plx_irqh(int irq, void *dev_id, struct pt_regs *regs) 
{
	u32 w1, w2;

	struct plx_dev *dev = (struct plx_dev *) dev_id;
	
	printk("%s default irq handler called at irq %d\n", dev->name,
	       irq);
	plx_read_config_dword(dev, 0x68, &w1);
	plx_read_config_dword(dev, 0x6c, &w2);

	printk("%s: INTCSR %08x INTCSR1 %08x", dev->name, w1, w2);
}

#if (LINUX_VERSION_CODE >= 0x020400)
/* 
 * Look for a free memory region of size bytes, size will be rounded up to
 * the next power of two. returns NULL on failure 
 */
unsigned long *find_mem_region(unsigned long *size)
{
	unsigned long base = 0x80000000;
	
	while (!(*size & base)) base /= 2;
     
	if ((base - 1) & *size) {
		printk("plx: 0x%08lx not a power of two. Setting to 0x%08lx\n",
		       *size, 2 * base);
		*size = 2 * base;
	}
	
	base = 0;

	do {
		if (check_mem_region(base, *size) < 0) {
			base -= *size;
		} else {
			return (unsigned long *) base;
		}
	} while (base > 0x40000000);
	
	return NULL;
}
#endif










